import contextlib
import functools
import io
import os
import pathlib
import random
import re
import shlex
import socket
import subprocess
import sys
import tarfile
import tempfile
import threading
import time
import urllib.request
import venv
from contextlib import contextmanager
from typing import Dict, List, Optional, Tuple

import click
import rich.console as rich_console
import rich.table
import validators
import watchfiles
from gql import gql

import fixieai.client
from fixieai import constants
from fixieai.cli.agent import agent_config
from fixieai.cli.agent import tunnel

# Regex pattern to match valid entry points: "module:object"
VAR_NAME_RE = r"(?![0-9])\w+"
ENTRY_POINT_PATTERN = re.compile(rf"^{VAR_NAME_RE}(\.{VAR_NAME_RE})*:{VAR_NAME_RE}$")
# The agent template main.py file
AGENT_TEMPLATE_URL = (
    "https://raw.githubusercontent.com/fixie-ai/fixie-examples/main/agents/template.py"
)

REQUIREMENTS_TXT = "requirements.txt"
CURRENT_FIXIE_REQUIREMENT = f"fixieai ~= {fixieai.__version__}"
FIXIE_REQUIREMENT_PATTERN = re.compile(r"^\s*fixieai([^\w]|$)")

# Environment variables for agents
FIXIE_ALLOWED_AGENT_ID = "FIXIE_ALLOWED_AGENT_ID"
FIXIE_REFRESH_ON_STARTUP = "FIXIE_REFRESH_ON_STARTUP"


@click.group(help="Agent-related commands.")
def agent():
    pass


def _slugify(s: str) -> str:
    s = s.lower().strip()
    s = re.sub(r"[^\w\s-]", "", s)
    s = re.sub(r"[\s_-]+", "-", s)
    s = s.strip("-")
    return s


def _cli_revision_metadata(command: str) -> dict:
    """Returns common agent revision metadata for CLI-created revisions."""
    return {
        "fixie.ai/client": "python-cli",
        "fixie.ai/python-cli/command": command,
        "fixie.ai/python-cli/version": fixieai.__version__,
    }


def _update_agent_requirements(
    existing_requirements: List[str], new_requirements: List[str]
) -> List[str]:
    """Returns an updated list of agent requirements to go in requirements.txt."""

    # If the user specified a fixieai requirement, use that.
    fixie_requirement = next(
        filter(
            functools.partial(re.match, FIXIE_REQUIREMENT_PATTERN), new_requirements
        ),
        CURRENT_FIXIE_REQUIREMENT,
    )

    # Ensure that a compatible version of the Fixie SDK is present.
    resolved_requirements: List[str] = []

    for existing_requirement in existing_requirements:
        if (
            re.match(FIXIE_REQUIREMENT_PATTERN, existing_requirement)
            and existing_requirement != fixie_requirement
        ):
            # Ignore any existing (but different) fixieai requirements.
            continue

        resolved_requirements.append(existing_requirement)

    for new_requirement in [fixie_requirement] + new_requirements:
        if new_requirement not in resolved_requirements:
            resolved_requirements.append(new_requirement)

    return resolved_requirements


@agent.command("init", help="Creates an agent.yaml file.")
@click.option("--handle")
@click.option("--description")
@click.option("--entry-point")
@click.option("--more-info-url")
@click.option(
    "--requirement",
    multiple=True,
    type=str,
    help="Additional requirements for requirements.txt. Can be specified multiple times.",
)
@click.argument("path", required=False)
def init_agent(path, handle, description, entry_point, more_info_url, requirement):
    # Create the directory if necessary.
    path = pathlib.Path(path or ".")
    path.mkdir(parents=True, exist_ok=True)

    try:
        current_config = agent_config.load_config(path)
    except FileNotFoundError:
        current_config = agent_config.AgentConfig(handle=_slugify(path.resolve().name))

    while handle is None or not validators.slug(handle):
        if handle is not None:
            click.secho("Handle can be alpha numerics, underscores and dashes only.")
        handle = click.prompt("Handle", default=current_config.handle)

    if description is None:
        description = click.prompt("Description", default=current_config.description)

    while entry_point is None or not ENTRY_POINT_PATTERN.match(entry_point):
        if entry_point is not None:
            click.echo(
                "Entrypoint must be in module:attribute format (e.g. 'main:agent', 'main:run')"
            )

        entry_point = click.prompt(
            "Python Entrypoint (module:attribute)", default=current_config.entry_point
        )

    while more_info_url is None or not validators.url(more_info_url):
        if more_info_url is not None:
            click.echo("More info URL must be a valid URL.")

        more_info_url = click.prompt(
            "More info URL", default=current_config.more_info_url
        )
        if not more_info_url:
            break

    current_config.handle = handle
    current_config.description = description
    current_config.entry_point = entry_point
    current_config.more_info_url = more_info_url
    agent_config.save_config(current_config, path)

    entry_module, _ = entry_point.split(":")
    expected_main_path = path / (entry_module.replace(".", "/") + ".py")
    if not os.path.exists(expected_main_path):
        urllib.request.urlretrieve(AGENT_TEMPLATE_URL, expected_main_path)
        click.secho(
            f"Initialized agent.yaml and made a template agent file at {expected_main_path}",
            fg="green",
        )
    else:
        click.secho(f"Initialized agent.yaml.", fg="green")

    try:
        with open(path / REQUIREMENTS_TXT, "rt") as requirements_txt:
            existing_requirements = list(
                r.strip() for r in requirements_txt.readlines()
            )
    except FileNotFoundError:
        existing_requirements = []

    resolved_requirements = _update_agent_requirements(
        existing_requirements, list(requirement)
    )
    if not existing_requirements:
        write_requirements = True
    else:
        new_requirements = [
            r for r in resolved_requirements if r not in existing_requirements
        ]
        removed_requirements = [
            r for r in existing_requirements if r not in resolved_requirements
        ]

        if new_requirements or removed_requirements:
            click.secho(
                f"{path / REQUIREMENTS_TXT} already exists.",
                fg="yellow",
            )
            if new_requirements:
                click.secho(
                    f"The following requirements will be added: {new_requirements}",
                    fg="yellow",
                )
            if removed_requirements:
                click.secho(
                    f"The following requirements will be removed: {removed_requirements}",
                    fg="yellow",
                )
            write_requirements = click.confirm("Okay to proceed?", default=True)
        else:
            write_requirements = False

    if write_requirements:
        with open(path / REQUIREMENTS_TXT, "wt") as requirements_txt:
            requirements_txt.writelines(r + "\n" for r in resolved_requirements)


@agent.command("list", help="List agents.")
@click.option("--verbose", is_flag=True, help="Enable verbose output.")
@click.pass_context
def list_agents(ctx, verbose):
    client = ctx.obj.client
    agents = client.get_agents()
    agent_ids = sorted(agents.keys())
    for agent_id in agent_ids:
        agent = agents[agent_id]
        click.secho(f"{agent_id}", fg="green", nl=False)
        click.echo(f": {agent['name']}")
        if verbose:
            click.echo(f"    {agent['description']}")
            if "moreInfoUrl" in agent and agent["moreInfoUrl"]:
                click.secho(f"    More info", fg="yellow", nl=False)
                click.echo(f": {agent['moreInfoUrl']}")
            click.echo()


@agent.command("show", help="Show an agent.")
@click.argument("agent_id")
@click.pass_context
def show_agent(ctx, agent_id: str):
    client = ctx.obj.client
    agent = client.get_agent(agent_id)
    if not agent.valid:
        click.echo(f"Agent {agent_id} not found.")
        return
    click.secho(f"{agent.agent_id}", fg="green", nl=False)
    click.echo(f"Owner: {agent.owner}")
    if agent.name:
        click.echo(f": {agent.name}")
    if agent.description:
        click.echo(f"Description: {agent.description}")
    if agent.more_info_url:
        click.echo(f"More info URL: {agent.more_info_url}")
    click.echo(f"Published: {agent.published}")
    click.echo(f"Created: {agent.created}")
    click.echo(f"Modified: {agent.modified}")
    if agent.queries:
        click.echo(f"Queries: {agent.queries}")


@agent.command("delete", help="Delete an agent.")
@click.argument("handle")
@click.pass_context
def delete_agent(ctx, handle: str):
    client = ctx.obj.client
    agent = client.get_agent(handle)
    agent.delete_agent()
    click.echo(f"Deleted agent {agent.agent_id}")


def _validate_agent_path(ctx, param, value):
    normalized = agent_config.normalize_path(value)
    if not os.path.exists(normalized):
        if not click.confirm(
            f"{normalized} does not exist. Would you like to create a new agent?",
            default=True,
        ):
            raise click.BadParameter(f"{normalized} does not exist")

        ctx.invoke(init_agent, path=value)

        # Re-normalize it after running init.
        normalized = agent_config.normalize_path(value)

    return normalized


def _ensure_agent_updated(
    client: fixieai.client.FixieClient,
    agent_id: str,
    config: agent_config.AgentConfig,
):
    agent = client.get_agent(agent_id)
    if not agent.valid:
        agent.create_agent(
            name=config.name,
            description=config.description,
            more_info_url=config.more_info_url,
            published=config.public or False,
        )
    else:
        agent.update_agent(
            new_handle=config.handle,
            name=config.name,
            description=config.description,
            more_info_url=config.more_info_url,
            published=config.public or False,
        )

    return agent


def _configure_venv(
    console: rich_console.Console, agent_dir: str
) -> Tuple[str, Dict[str, str]]:
    """Configures a virtual environment for the agent with its Python dependencies installed."""

    venv_path = os.path.join(agent_dir, ".fixie.venv")
    console.print(f"Configuring virtual environment in {venv_path}")

    is_windows = os.name == "nt"
    bin_path = os.path.join(venv_path, "bin" if not is_windows else "Scripts")
    python_exe = os.path.join(bin_path, "python")
    if not os.path.exists(python_exe):
        # Match `python -m venv` behavior with symlinks on non-Windows to work around https://bugs.python.org/issue38705
        venv.create(venv_path, with_pip=True, symlinks=not is_windows)
    requirements_txt_path = os.path.join(agent_dir, REQUIREMENTS_TXT)
    if not os.path.exists(requirements_txt_path):
        raise ValueError(
            f"There is no {REQUIREMENTS_TXT} file at {requirements_txt_path}, please re-run [bold]fixie init[/bold]."
        )
    subprocess.run(
        [
            python_exe,
            "-m",
            "pip",
            "install",
            "-r",
            requirements_txt_path,
            "--disable-pip-version-check",
        ]
    ).check_returncode()
    console.print(f"[green]Requirements installed successfully.[/green]")

    env = os.environ.copy()

    # Quasi-activate the venv by putting the bin directory at the front of the path
    env["PATH"] = bin_path + ":" + env.get("PATH", "")

    return python_exe, env


def _validate_agent_loads_or_exit(
    console: rich_console.Console,
    agent_dir: str,
    agent_handle: str,
    python_exe: str,
    agent_env: Dict[str, str],
):
    """Validates that an agent can successfully load and exits the current process if it can't."""
    loader_process = subprocess.run(
        [os.path.abspath(python_exe), "-m", "fixieai.cli.agent.loader"],
        env=agent_env,
        cwd=agent_dir or None,
    )
    if loader_process.returncode:
        console.print(
            f'[red]Agent "{agent_handle}" could not be loaded. If your agent requires additional dependencies '
            f"make sure they are specified in {os.path.join(agent_dir, REQUIREMENTS_TXT)}.[/red]"
        )
        sys.exit(loader_process.returncode)


@contextlib.contextmanager
def _temporary_revision_replacer(
    console: rich.console.Console, client: fixieai.FixieClient, agent_handle: str
):
    """Yields a function that can be used to temporarily replace an agent's current revision."""

    # Get the preexisting revision so that we can restore it when we're done.
    existing_revision_id = client.get_current_agent_revision(agent_handle)
    current_temporary_revision: Optional[str] = None

    if existing_revision_id is not None:
        console.print(
            f"[yellow]This temporarily replaces existing revision {existing_revision_id}.[/yellow]"
        )

    def replace_revision(external_url: str):
        nonlocal current_temporary_revision

        # Create a new temporary revision.
        revision_to_delete = current_temporary_revision
        current_temporary_revision = client.create_agent_revision(
            handle=agent_handle,
            make_current=True,
            metadata=_cli_revision_metadata("serve"),
            external_url=external_url,
        )

        # Delete the previous temporary revision.
        if revision_to_delete:
            client.delete_agent_revision(agent_handle, revision_to_delete)

        console.print(
            f"[green]Now serving temporary revision {current_temporary_revision}.[/green]"
        )

    try:
        yield replace_revision
    finally:
        if existing_revision_id:
            client.set_current_agent_revision(agent_handle, existing_revision_id)
            console.print(
                f"[yellow]Restored existing revision {existing_revision_id}.[/yellow]"
            )

        if current_temporary_revision:
            client.delete_agent_revision(agent_handle, current_temporary_revision)


@contextlib.contextmanager
def _server_process(
    console: rich_console.Console,
    on_exit_event: threading.Event,
    python_exe: str,
    host: str,
    port: int,
    agent_env: Dict[str, str],
    agent_dir: str,
    startup_timeout_seconds: float,
):
    """Yields a Popen object with a running and ready server process."""

    with subprocess.Popen(
        [
            os.path.abspath(python_exe),
            "-m",
            "uvicorn",
            "fixieai.cli.agent.loader:uvicorn_app_factory",
            "--host",
            host,
            "--port",
            str(port),
            "--factory",
        ],
        env=agent_env,
        cwd=agent_dir or None,
    ) as popen:

        def _wait_for_exit():
            popen.wait()
            on_exit_event.set()

        waiter_thread = threading.Thread(target=_wait_for_exit, daemon=True)
        waiter_thread.start()

        try:
            # Wait for the server to start up by polling the port.
            deadline = time.monotonic() + startup_timeout_seconds
            while time.monotonic() < deadline and popen.poll() is None:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    try:
                        s.connect(("127.0.0.1" if host == "0.0.0.0" else host, port))
                        break
                    except ConnectionRefusedError:
                        time.sleep(0.1)

            if time.monotonic() >= deadline or popen.poll() is not None:
                console.print("Server process failed to start.", style="red")
                raise click.Abort()

            yield popen
        finally:
            popen.terminate()
            waiter_thread.join()


def _wait_for_agent_changes_or_event(
    agent_dir: str, event: threading.Event
) -> Optional[str]:
    """Waits for any changes in the specified directory or the event to be set."""

    for changes in watchfiles.watch(
        agent_dir,
        watch_filter=lambda change, path: path.endswith(".py"),
        stop_event=event,
    ):
        for change, path in changes:
            return path

    return None


@agent.command(
    "serve", help="Serve the current agent locally via a publicly-accessible URL."
)
@click.argument("path", callback=_validate_agent_path, required=False)
@click.option("--host", default="0.0.0.0")
@click.option("--port", type=int, default=8181)
@click.option(
    "--tunnel/--no-tunnel",
    "use_tunnel",
    is_flag=True,
    default=True,
    help="(default enabled) Create a tunnel using localhost.run.",
)
@click.option(
    "--reload/--no-reload",
    is_flag=True,
    default=True,
    help="(default enabled) Reload automatically.",
)
@click.option(
    "--venv/--no-venv",
    "use_venv",
    is_flag=True,
    default=False,
    help="Run from virtual environment",
)
@click.pass_context
def serve(ctx, path, host, port, use_tunnel, reload, use_venv):
    console = rich_console.Console(soft_wrap=True)
    config = agent_config.load_config(path)
    client: fixieai.FixieClient = ctx.obj.client
    agent_id = f"{client.get_current_username()}/{config.handle}"

    with contextlib.ExitStack() as stack:
        agent_dir = os.path.dirname(path) or "."

        # Configure the virtual enviroment
        if use_venv:
            python_exe, agent_env = _configure_venv(console, agent_dir)
        else:
            python_exe, agent_env = sys.executable, os.environ.copy()
        agent_env[FIXIE_ALLOWED_AGENT_ID] = agent_id

        if use_tunnel:
            # Start up a tunnel via localhost.run.
            console.print(f"Opening tunnel to {host}:{port} via localhost.run.")
            deployment_urls_iter = iter(stack.enter_context(tunnel.Tunnel(port)))
        elif config.deployment_url:
            deployment_urls_iter = iter([config.deployment_url])
        else:
            console.print(
                f"[red]You must specify a deployment_url in agent.yaml or use --tunnel.[/red]"
            )
            raise click.Abort()

        agent_api = _ensure_agent_updated(client, agent_id, config)
        console.print(
            f"You can access your agent via [cyan]{client.get_agent_page_url(agent_api.agent_id)}[/]"
            f" or `[green]fixie console --agent {agent_api.agent_id}[/]`."
        )

        replace_revision = stack.enter_context(
            _temporary_revision_replacer(console, client, config.handle)
        )

        # An Event that indicates that either the subprocess exited or the tunnel URL has been rotated.
        requires_action = threading.Event()
        current_url = next(deployment_urls_iter)

        def _watch_tunnel():
            """Watches the tunnel for URL rotations."""
            nonlocal current_url

            while True:
                try:
                    current_url = next(deployment_urls_iter)
                except StopIteration:
                    return
                finally:
                    requires_action.set()

        threading.Thread(target=_watch_tunnel, daemon=True).start()

        while True:
            requires_action.clear()

            with _server_process(
                console,
                requires_action,
                python_exe,
                host,
                port,
                agent_env,
                agent_dir,
                startup_timeout_seconds=10.0,
            ) as popen:

                # The server has started up, so we can replace the revision.
                replace_revision(current_url)

                while True:
                    if reload:
                        changed_file = _wait_for_agent_changes_or_event(
                            agent_dir, requires_action
                        )
                        if changed_file is not None:
                            console.print(
                                f"[yellow]{changed_file} was changed. Restarting server.[/yellow]"
                            )
                            break
                        else:
                            assert requires_action.is_set()

                    requires_action.wait()
                    requires_action.clear()

                    if popen.poll() is not None:
                        console.print(
                            "The server process stopped, exiting.",
                            style="red" if popen.returncode != 0 else None,
                        )
                        raise click.exceptions.Exit(popen.returncode)
                    else:
                        # The tunnel URL changed.
                        replace_revision(current_url)


_DEPLOYMENT_BOOTSTRAP_SOURCE = """
import os

import dotenv

if __name__ == "__main__":
    dotenv.load_dotenv()
    os.chdir("agent")

    # N.B. Load the .dotenv file before loading the Fixie SDK
    import uvicorn

    from fixieai.cli.agent import loader
    port = int(os.getenv("PORT", 8080))
    uvicorn.run(loader.uvicorn_app_factory(), host="0.0.0.0", port=port)
"""


@contextmanager
def _spinner(console: rich_console.Console, text: str):
    """A context manager that renders a spinner and ✅/❌ on completion."""

    try:
        with console.status(text):
            yield
        console.print(f":white_check_mark: {text}")
    except:
        console.print(f":cross_mark: {text}")
        raise


def _tarinfo_filter(
    console: rich_console.Console, root: str, tar_prefix: str, tarinfo: tarfile.TarInfo
) -> Optional[tarfile.TarInfo]:
    """Filters out hidden files from being included in a tarball."""

    basename = os.path.basename(tarinfo.name)
    if basename.startswith(".") and basename != ".env":
        original_path = tarinfo.name
        if original_path.startswith(tar_prefix):
            original_path = original_path[len(tar_prefix) :]

        original_path = os.path.join(root, original_path)
        console.print(
            f"Ignoring hidden {'directory' if tarinfo.isdir() else 'file'} {original_path}",
            style="grey53",
        )
        return None
    else:
        return tarinfo


def _add_text_file_to_tarfile(path: str, text: str, tarball: tarfile.TarFile):
    file_bytes = text.encode("utf-8")
    tarinfo = tarfile.TarInfo(path)
    tarinfo.size = len(file_bytes)
    tarball.addfile(tarinfo, io.BytesIO(file_bytes))


@contextlib.contextmanager
def _agent_code_package(agent_dir: str, agent_id: str, console: rich_console.Console):
    """Yields a file-like code package that can be deployed to Fixie."""

    with tempfile.TemporaryFile() as tarball_file:
        with tarfile.open(fileobj=tarball_file, mode="w:gz") as tarball:
            tarball.add(
                agent_dir,
                arcname="agent",
                filter=functools.partial(_tarinfo_filter, console, agent_dir, "agent/"),
            )

            # Add the bootstrapping code and environment variables
            _add_text_file_to_tarfile("main.py", _DEPLOYMENT_BOOTSTRAP_SOURCE, tarball)
            _add_text_file_to_tarfile(
                ".env",
                "\n".join(
                    f"{key}={value}"
                    for key, value in {
                        FIXIE_ALLOWED_AGENT_ID: agent_id,
                        "FIXIE_API_URL": constants.FIXIE_API_URL,
                    }.items()
                ),
                tarball,
            )

        tarball_file.seek(0)
        yield tarball_file


@agent.command("deploy", help="Deploy the current agent.")
@click.argument("path", callback=_validate_agent_path, required=False)
@click.option(
    "--metadata-only",
    is_flag=True,
    default=None,
    help="(No longer supported.)",
)
@click.option("--public", is_flag=True, help="Make the agent public.", default=None)
@click.option(
    "--validate/--no-validate",
    "validate",
    is_flag=True,
    default=True,
    help="(default enabled) Validate that the agent loads in a local venv before deploying.",
)
@click.option(
    "--current/--not-current",
    "make_current",
    is_flag=True,
    default=True,
    help="(default enabled) Change the current revision to the deployed revision.",
)
@click.option(
    "--metadata",
    nargs=2,
    type=str,
    help="Additional metadata to associate with the revision.",
    multiple=True,
)
@click.option(
    "--reindex/--no-reindex",
    "reindex",
    is_flag=True,
    default=False,
    help="Force any corpora to be reindexed.",
)
@click.pass_context
def deploy(ctx, path, metadata_only, public, validate, make_current, metadata, reindex):
    console = rich_console.Console(soft_wrap=True)
    if metadata_only is not None:
        console.print(
            "[yellow]Warning:[/] The [blue]--metadata-only[/] option is no longer supported."
        )

    config = agent_config.load_config(path)
    if config.public is not None:
        console.print(
            "[yellow]Warning:[/] The [blue]`public`[/] option in the agent config is deprecated and will be removed soon."
        )
        console.print(
            "[yellow]Warning:[/] Use [blue]`fixie agent deploy --public`[/] instead."
        )
    if public:
        config.public = True

    client: fixieai.FixieClient = ctx.obj.client
    agent_id = f"{client.get_current_username()}/{config.handle}"
    console.print(f"🦊 Deploying agent [green]{agent_id}[/]...")

    agent_dir = os.path.dirname(path) or "."
    if validate:
        # Validate that the agent loads in a virtual environment before deploying.
        python_exe, agent_env = _configure_venv(console, agent_dir)
        agent_env[FIXIE_ALLOWED_AGENT_ID] = agent_id
        _validate_agent_loads_or_exit(
            console, agent_dir, config.handle, python_exe, agent_env
        )

    _ = _ensure_agent_updated(client, agent_id, config)

    metadata_dict = {**dict(metadata or []), **_cli_revision_metadata("deploy")}

    if config.deployment_url is None:
        # Deploy the code to Fixie.
        with _agent_code_package(
            agent_dir, agent_id, console
        ) as tarball_file, _spinner(console, "Deploying..."):
            revision_id = client.create_agent_revision(
                config.handle,
                make_current=make_current,
                metadata=metadata_dict,
                python_gzip_tarfile=tarball_file,
                reindex_corpora=reindex,
            )
    else:
        # Create a new revision with the specified URL.
        with _spinner(console, "Deploying..."):
            revision_id = client.create_agent_revision(
                config.handle,
                make_current=make_current,
                metadata=metadata_dict,
                external_url=config.deployment_url,
                reindex_corpora=reindex,
            )

    sample_queries = client.gqlclient.execute(
        gql(
            """
            query GetSampleQueries($agentId: String!) {
                agentById(agentId: $agentId) {
                    queries
                }
            }
            """
        ),
        variable_values={"agentId": agent_id},
    )

    if sample_queries["agentById"]["queries"]:
        suggested_query = random.choice(sample_queries["agentById"]["queries"])
    else:
        suggested_query = "Hello!"

    console.print(
        f"Revision {revision_id} was deployed to {constants.FIXIE_API_URL}/agents/{agent_id}\n"
        f"You can also chat with your agent using the fixie CLI:\n\n"
        f"fixie console --agent {agent_id} {shlex.quote(suggested_query)}"
    )


@agent.command("publish", help="Make the current agent public.")
@click.argument("path", callback=_validate_agent_path, required=False)
@click.pass_context
def publish(ctx, path):
    console = rich_console.Console(soft_wrap=True)
    config = agent_config.load_config(path)
    if config.public is not None:
        console.print(
            "[yellow]Warning:[/] The [blue]`public`[/] option in the agent config is deprecated and will be removed soon."
        )
    client: fixieai.FixieClient = ctx.obj.client
    agent_id = f"{client.get_current_username()}/{config.handle}"
    console.print(f"🦊 Publishing agent [green]{agent_id}[/]...")

    agent = client.get_agent(agent_id)
    if not agent.valid:
        console.print(
            f"[red]Error:[/] Agent [green]{agent_id}[/] not deployed. Use [blue]`fixie agent deploy`[/] to deploy it."
        )
        raise click.Abort()
    agent.update_agent(published=True)

    console.print(f"Agent [green]{agent_id}[/] has been made public.")


@agent.command("unpublish", help="Make the current agent not public.")
@click.argument("path", callback=_validate_agent_path, required=False)
@click.pass_context
def unpublish(ctx, path):
    console = rich_console.Console(soft_wrap=True)
    config = agent_config.load_config(path)
    if config.public is not None:
        console.print(
            "[yellow]Warning:[/] The [blue]`public`[/] option in the agent config is deprecated and will be removed soon."
        )
    client: fixieai.FixieClient = ctx.obj.client
    agent_id = f"{client.get_current_username()}/{config.handle}"
    console.print(f"🦊 Unpublishing agent [green]{agent_id}[/]...")

    agent = client.get_agent(agent_id)
    if not agent.valid:
        console.print(
            f"[red]Error:[/] Agent [green]{agent_id}[/] not deployed. Use [blue]`fixie agent deploy`[/] to deploy it."
        )
        raise click.Abort()
    agent.update_agent(published=False)

    console.print(f"Agent [green]{agent_id}[/] has been made private.")


@agent.group(help="Revisions-related commands.")
def revisions():
    pass


@revisions.command("list", help="Lists the revisions associated with an agent.")
@click.argument("path", callback=_validate_agent_path, required=False)
@click.option("--json", is_flag=True, default=False)
@click.pass_context
def list_revisions(ctx, path, json):
    console = rich_console.Console(soft_wrap=True)
    config = agent_config.load_config(path)

    client: fixieai.FixieClient = ctx.obj.client
    agent_id = f"{client.get_current_username()}/{config.handle}"
    agent = client.get_agent(agent_id)
    if not agent.valid:
        console.print(
            f"[red]Error:[/] Agent [green]{agent_id}[/] not deployed. Use [blue]`fixie agent deploy`[/] to deploy it."
        )
        raise click.Abort()

    query = gql(
        """
        query GetRevisions($agentId: String!) {
            agentById(agentId: $agentId) {
                allRevisions {
                    id
                    created
                    isCurrent
                    indexedCorpora {
                        status
                        created
                        modified
                        vectorCount
                    }
                    deployment {
                        __typename
                        ... on ExternalDeployment {
                            url
                        }
                        ... on ManagedDeployment {
                            environment
                        }
                    }
                    metadata {
                        key
                        value
                    }
                }
            }
        }
        """
    )
    revisions = client.gqlclient.execute(query, variable_values={"agentId": agent_id})[
        "agentById"
    ]["allRevisions"]

    if json:
        console.print_json(data=revisions)
    else:
        table = rich.table.Table(title=f"{agent_id} revisions")
        table.add_column("ID", no_wrap=True)
        table.add_column("Date")
        table.add_column("Deployment")

        has_corpora = False
        if any(revision["indexedCorpora"] for revision in revisions):
            table.add_column("Indexed")
            has_corpora = True

        table.add_column("Current")

        for revision in revisions:
            row = [
                revision["id"],
                revision["created"],
            ]

            deployment = revision["deployment"]
            if deployment is None:
                row.append("-")
            elif deployment["__typename"] == "ExternalDeployment":
                row.append(deployment["url"])
            elif deployment["__typename"] == "ManagedDeployment":
                row.append(f"Fixie-hosted {deployment['environment'].title()}")
            else:
                row.append("?")

            if has_corpora:
                indexes = revision["indexedCorpora"] or []
                if any(index["status"] == "FAILED" for index in indexes):
                    row.append("❌")
                elif any(index["status"] == "IN_PROGRESS" for index in indexes):
                    row.append("⏳")
                elif indexes and all(
                    index["status"] == "COMPLETED" for index in indexes
                ):
                    row.append("✅")
                else:
                    row.append("")

            row.append("✅" if revision["isCurrent"] else "")

            table.add_row(*row)

        console.print(table)


@revisions.command(
    "set-current", help="Sets the current (active) revision of an agent."
)
@click.argument("path", callback=_validate_agent_path, required=False)
@click.option("--id", required=True, help="The revision ID to make current.")
@click.pass_context
def set_current_revision(ctx, path, id):
    console = rich_console.Console(soft_wrap=True)
    config = agent_config.load_config(path)

    client: fixieai.FixieClient = ctx.obj.client
    client.set_current_agent_revision(config.handle, id)
    console.print(f"Set current revision to {id}.")


@revisions.command("delete", help="Deletes a revision of an agent.")
@click.argument("path", callback=_validate_agent_path, required=False)
@click.option("--id", required=True, help="The revision ID to delete.")
@click.option("--force/-f", help="Delete the revision even if it is the current one.")
@click.pass_context
def delete_revision(ctx, path, id, force):
    console = rich_console.Console(soft_wrap=True)
    config = agent_config.load_config(path)

    client: fixieai.FixieClient = ctx.obj.client

    current_revision = client.get_current_agent_revision(config.handle)
    if current_revision == id:
        if not force:
            console.print(
                f"Not deleting revision {id} because it is the current revision. (Use --force to delete it anyway.)",
                style="red",
            )
            raise click.exceptions.Exit(1)
        else:
            console.print(
                f"Deleting current revision {id} because --force was set. This will prevent the agent from responding to requests.",
                style="yellow",
            )

    client.delete_agent_revision(config.handle, id)
    console.print(f"Deleted revision {id}.")



import pytest

from fixieai import constants
from fixieai.client.client import FixieClient


@pytest.fixture
def fixie_client():
    return FixieClient(api_key="test-key")


def test_get_agents(fixie_client, requests_mock):
    requests_mock.post(
        constants.FIXIE_GRAPHQL_URL,
        json={
            "data": {"allAgents": [{"agentId": "testuser/test", "name": "Test Agent"}]}
        },
    )
    assert fixie_client.get_agents() == {
        "testuser/test": {"agentId": "testuser/test", "name": "Test Agent"}
    }


def test_get_agent(fixie_client, requests_mock):
    requests_mock.post(
        constants.FIXIE_GRAPHQL_URL,
        json={
            "data": {
                "agentById": {
                    "agentId": "testuser/test-agent-handle",
                    "handle": "test-agent-handle",
                    "name": "Test Agent",
                    "description": "Test Agent Description",
                    "queries": ["test-query"],
                    "moreInfoUrl": "https://test.com",
                    "published": True,
                    "owner": {
                        "username": "testuser",
                    },
                }
            }
        },
    )
    agent = fixie_client.get_agent("testuser/test-agent-handle")
    assert agent.valid
    assert agent.handle == "test-agent-handle"
    assert agent.agent_id == "testuser/test-agent-handle"
    assert agent.name == "Test Agent"
    assert agent.description == "Test Agent Description"
    assert agent.queries == ["test-query"]
    assert agent.more_info_url == "https://test.com"
    assert agent.published is True
    assert agent.owner == "testuser"


def test_create_agent(fixie_client, requests_mock):
    requests_mock.post(
        constants.FIXIE_GRAPHQL_URL,
        json={
            "data": {},
            "errors": ["GraphQL error: No such agent: testuser/test-agent-handle"],
        },
    )
    agent = fixie_client.get_agent("testuser/test-agent-handle")
    assert agent.handle == "test-agent-handle"
    assert agent.agent_id == "testuser/test-agent-handle"
    assert not agent.valid

    # There are two GraphQL calls resulting from create_agent, we need
    # to mock both of the responses.
    requests_mock.post(
        constants.FIXIE_GRAPHQL_URL,
        [
            {
                "json": {
                    "data": {
                        "createAgent": {
                            "agent": {
                                "agentId": "testuser/test-agent-handle",
                            }
                        }
                    }
                },
            },
            {
                "json": {
                    "data": {
                        "agentById": {
                            "agentId": "testuser/test-agent-handle",
                            "handle": "test-agent-handle",
                            "name": "Test Agent",
                            "description": "Test Agent Description",
                            "queries": [],
                            "moreInfoUrl": "https://test.com",
                            "published": True,
                            "owner": {
                                "username": "testuser",
                            },
                        }
                    }
                },
            },
        ],
    )
    agent.create_agent(
        name="Test Agent",
        description="Test Agent Description",
        more_info_url="https://test.com",
        published=True,
    )
    assert agent.valid
    assert agent.agent_id == "testuser/test-agent-handle"
    assert agent.name == "Test Agent"
    assert agent.description == "Test Agent Description"
    assert agent.queries == []
    assert agent.more_info_url == "https://test.com"
    assert agent.published is True
    assert agent.owner == "testuser"


def test_sessions(fixie_client, requests_mock):
    requests_mock.post(
        constants.FIXIE_GRAPHQL_URL,
        json={
            "data": {"allSessions": [{"handle": "test-handle", "name": "Test Session"}]}
        },
    )
    assert fixie_client.get_sessions() == ["test-handle"]


def test_get_session(fixie_client, requests_mock):
    requests_mock.post(
        constants.FIXIE_GRAPHQL_URL,
        json={
            "data": {
                "sessionByHandle": {"handle": "test-handle", "frontendAgentId": None}
            }
        },
    )
    session = fixie_client.get_session("test-handle")

    requests_mock.post(
        constants.FIXIE_GRAPHQL_URL,
        json={
            "data": {
                "sessionByHandle": {
                    "handle": "test-handle",
                    "messages": [
                        {
                            "id": 1,
                            "text": "Test message",
                        }
                    ],
                }
            }
        },
    )
    assert session.get_messages() == [{"id": 1, "text": "Test message"}]



import dataclasses
from typing import Any, MutableMapping, TextIO, Type, TypeVar, Union

import dataclasses_json
import yaml

A = TypeVar("A", bound="DataClassYamlMixin")


class _NicelyFormattedYAMLString:
    """A string wrapper that formats multiline strings more nicely."""

    def __init__(self, value: str):
        self.value = value

    def dump(self, dumper: yaml.Dumper) -> yaml.Node:
        return dumper.represent_scalar(
            "tag:yaml.org,2002:str", self.value, style="|" if "\n" in self.value else ""
        )


yaml.add_representer(
    _NicelyFormattedYAMLString,
    lambda dumper, value: value.dump(dumper),
)


class DataClassYamlMixin(dataclasses_json.DataClassJsonMixin):
    @classmethod
    def from_yaml(cls: Type[A], config: Union[str, TextIO]) -> A:
        return cls.from_dict(yaml.safe_load(config) or {})

    def to_yaml(self) -> str:
        as_dict: MutableMapping[str, Any] = self.to_dict()

        # Exclude any keys whose values and default values are None.
        for field in dataclasses.fields(self):
            if field.name not in as_dict:
                # Not all fields serialize to their Python names, leave them as-is.
                continue

            value = as_dict[field.name]
            if value is None and field.default is None:
                del as_dict[field.name]
            elif isinstance(value, str):
                as_dict[field.name] = _NicelyFormattedYAMLString(value)

        return yaml.dump(as_dict, sort_keys=False)  # type: ignore



import os
import pathlib
import queue
import random
import string
import threading
import time
import unittest.mock
from typing import Callable, Dict, Optional

import click.testing
import gql.client
import pytest
import validators

import fixieai
from fixieai import constants
from fixieai.cli import cli
from fixieai.cli.agent import agent_config

from . import commands


@pytest.fixture(autouse=True)
def mock_fixie_api_key(mocker):
    mocker.patch.object(constants, "fixie_api_key", return_value="TEST")
    os.environ["FIXIE_API_KEY"] = "TEST"


@pytest.fixture(autouse=True)
def mock_graphql(mocker):
    response_mocks: Dict[str, Callable] = {}

    def _execute(document, *args, **kwargs):
        mocked_queries = {
            "getUsername": {"user": {"username": "testuser"}},
            "getAgentById": {
                "agentById": {
                    "agentId": "testuser/testagent",
                    "handle": "testagent",
                    "name": "Test Agent",
                    "description": "Test agent description",
                    "queries": [],
                    "funcUrl": "https://example.com",
                    "moreInfoUrl": "https://example.com",
                    "published": False,
                    "owner": {
                        "username": "testuser",
                    },
                }
            },
            "UpdateAgent": {
                "updateAgent": {
                    "agent": {
                        "agentId": "testuser/testagent",
                    }
                }
            },
            "SetCurrentAgentRevision": {
                "updateAgent": {
                    "agent": {
                        "agentId": "testuser/testagent",
                    }
                }
            },
            "CreateAgentRevision": {
                "createAgentRevision": {"revision": {"id": "FAKEID"}}
            },
            "GetSampleQueries": {
                "agentById": {"queries": ["Why did the chicken cross the road?"]}
            },
            "GetRevisionId": {"agentById": {"currentRevision": {"id": "FAKEID"}}},
            "DeleteRevision": {
                "deleteAgentRevision": {"agent": {"agentId": "testuser/testagent"}}
            },
        }

        operation = document.definitions[0].name.value
        if operation in response_mocks:
            return response_mocks[operation](document, *args, **kwargs)
        if operation in mocked_queries:
            return mocked_queries[operation]
        else:
            raise ValueError("Unmocked GraphQL operation: " + operation)

    mocker.patch.object(gql.client.Client, "execute", side_effect=_execute)
    return response_mocks


@pytest.fixture(autouse=True)
def mock_tunnel(mocker):
    tunnel = mocker.patch("fixieai.cli.agent.tunnel.Tunnel")
    tunnel.return_value.__enter__.return_value = ["mock://mock.example.com"]
    return tunnel


@pytest.fixture(autouse=True)
def mock_fixie_requirement(mocker):
    # Use the current version of fixieai as the fixie dependency.
    mocker.patch.object(
        commands,
        "CURRENT_FIXIE_REQUIREMENT",
        str(pathlib.Path(fixieai.__file__).parent.parent),
    )


@pytest.fixture()
def mock_server_process_ctx(mocker):
    mock_context_manager = unittest.mock.MagicMock()
    mock_context_manager.stop_queue = queue.Queue()

    def _server_process(console, event, *_, **__):
        mock_process = unittest.mock.MagicMock()
        mock_process.poll.return_value = None

        def _stop(code):
            mock_process.poll.return_value = code
            mock_process.returncode = code
            event.set()

        def _enter(*args, **kwargs):
            if mock_context_manager.stop_immediately == True:
                _stop(0)
            else:
                mock_context_manager.stop_queue.put(_stop)

            return mock_process

        mock_context_manager.__enter__.side_effect = _enter
        return mock_context_manager

    mocker.patch.object(commands, "_server_process", side_effect=_server_process)

    return mock_context_manager


def test_update_agent_requirements_adds_fixie():
    new_requirements = commands._update_agent_requirements([], [])
    assert new_requirements == [commands.CURRENT_FIXIE_REQUIREMENT]


def test_update_agent_requirements_honors_specified_fixie():
    new_requirements = commands._update_agent_requirements([], ["fixieai"])
    assert new_requirements == ["fixieai"]


def test_update_agent_requirements_overrides_existing_fixie():
    new_requirements = commands._update_agent_requirements(
        ["fixieai", "fixieai == 0.0"], []
    )
    assert new_requirements == [commands.CURRENT_FIXIE_REQUIREMENT]


def test_update_agent_requirements_combines():
    new_requirements = commands._update_agent_requirements(
        ["package1", "package2"], ["package2", "package3"]
    )
    assert new_requirements == [
        "package1",
        "package2",
        commands.CURRENT_FIXIE_REQUIREMENT,
        "package3",
    ]


@pytest.mark.parametrize("seed", range(100))
def test_slugify(seed):
    random.seed(seed)

    text = "".join(random.choices(string.ascii_letters, k=40))
    slugified = commands._slugify(text)
    assert len(slugified) <= len(text)
    assert validators.slug(slugified)


def test_init():
    runner = click.testing.CliRunner()
    with runner.isolated_filesystem() as temp_dir:
        result = runner.invoke(commands.init_agent)
        assert result.exit_code == 0

        config = agent_config.load_config(temp_dir)
        assert config.handle == commands._slugify(os.path.basename(temp_dir))
        assert os.path.exists(os.path.join(temp_dir, "main.py"))
        assert os.path.exists(os.path.join(temp_dir, "requirements.txt"))


def test_init_with_path():
    runner = click.testing.CliRunner()
    with runner.isolated_filesystem() as temp_dir:
        result = runner.invoke(commands.init_agent, ["foo/bar/baz"])
        assert result.exit_code == 0

        agent_dir = os.path.join(temp_dir, "foo/bar/baz")
        config = agent_config.load_config(agent_dir)
        assert config.handle == "baz"
        assert os.path.exists(os.path.join(agent_dir, "main.py"))
        assert os.path.exists(os.path.join(agent_dir, "requirements.txt"))


def test_init_args():
    runner = click.testing.CliRunner()
    with runner.isolated_filesystem() as temp_dir:
        result = runner.invoke(
            commands.init_agent,
            [
                "foo",
                "--handle",
                "my-handle",
                "--description",
                "a description",
                "--entry-point",
                "main:entrypoint",
                "--more-info-url",
                "https://example.com",
            ],
        )
        assert result.exit_code == 0
        agent_dir = os.path.join(temp_dir, "foo")
        config = agent_config.load_config(agent_dir)
        assert config.handle == "my-handle"
        assert config.description == "a description"
        assert config.entry_point == "main:entrypoint"
        assert config.more_info_url == "https://example.com"


def test_init_validation_prompting():
    runner = click.testing.CliRunner()
    with runner.isolated_filesystem() as temp_dir:
        result = runner.invoke(
            commands.init_agent,
            ["foo"],
            input="""my invalid handle
my-handle
a description
invalid entry point
main:entrypoint
invalid URL

""",
        )
        assert result.exit_code == 0
        agent_dir = os.path.join(temp_dir, "foo")
        config = agent_config.load_config(agent_dir)
        assert config.handle == "my-handle"
        assert config.description == "a description"
        assert config.entry_point == "main:entrypoint"
        assert config.more_info_url == ""


def test_temporary_revision_replacer(mock_graphql):
    """Tests the _temporary_revision_replacer context manager."""

    mock_graphql["GetRevisionId"] = unittest.mock.MagicMock()
    mock_graphql["CreateAgentRevision"] = unittest.mock.MagicMock()
    mock_graphql["DeleteRevision"] = unittest.mock.MagicMock()
    mock_graphql["SetCurrentAgentRevision"] = unittest.mock.MagicMock()

    counter = 0

    def next_revision_id():
        nonlocal counter
        counter += 1
        return str(counter)

    mock_graphql["GetRevisionId"].side_effect = lambda *_, **__: {
        "agentById": {"currentRevision": {"id": str(counter)}}
    }
    mock_graphql["CreateAgentRevision"].side_effect = lambda *_, **__: {
        "createAgentRevision": {"revision": {"id": next_revision_id()}}
    }

    mock_console = unittest.mock.MagicMock()
    with commands._temporary_revision_replacer(
        mock_console, fixieai.FixieClient(), "testagent"
    ) as replacer:
        # The first replacement should create a new revision.
        assert mock_graphql["CreateAgentRevision"].call_count == 0
        replacer("mock://mock.example.org")
        assert (
            mock_graphql["CreateAgentRevision"].call_args[1]["variable_values"][
                "externalDeployment"
            ]["url"]
            == "mock://mock.example.org"
        )
        assert mock_graphql["DeleteRevision"].call_count == 0

        # The second replacement should create a new revision and delete the previously created revision.
        replacer("mock://mock2.example.org")
        assert (
            mock_graphql["CreateAgentRevision"].call_args[1]["variable_values"][
                "externalDeployment"
            ]["url"]
            == "mock://mock2.example.org"
        )
        assert (
            mock_graphql["DeleteRevision"].call_args[1]["variable_values"]["revisionId"]
            == "1"
        )

    # When the context manager exits it should delete the last temporary revision and restore the initial revision.
    assert (
        mock_graphql["DeleteRevision"].call_args[1]["variable_values"]["revisionId"]
        == "2"
    )
    assert (
        mock_graphql["SetCurrentAgentRevision"].call_args[1]["variable_values"][
            "currentRevisionId"
        ]
        == "0"
    )


def test_temporary_revision_replacer_no_current_revision(mock_graphql):
    """Tests the _temporary_revision_replacer context manager when there is no current revision."""

    mock_graphql["GetRevisionId"] = unittest.mock.MagicMock()
    mock_graphql["CreateAgentRevision"] = unittest.mock.MagicMock()
    mock_graphql["DeleteRevision"] = unittest.mock.MagicMock()
    mock_graphql["SetCurrentAgentRevision"] = unittest.mock.MagicMock()

    counter = 0

    def next_revision_id():
        nonlocal counter
        counter += 1
        return str(counter)

    mock_graphql["GetRevisionId"].return_value = {
        "agentById": {"currentRevision": None}
    }
    mock_graphql["CreateAgentRevision"].side_effect = lambda *_, **__: {
        "createAgentRevision": {"revision": {"id": next_revision_id()}}
    }

    mock_console = unittest.mock.MagicMock()
    with commands._temporary_revision_replacer(
        mock_console, fixieai.FixieClient(), "testagent"
    ) as replacer:
        assert mock_graphql["CreateAgentRevision"].call_count == 0
        replacer("mock://mock.example.org")
        assert (
            mock_graphql["CreateAgentRevision"].call_args[1]["variable_values"][
                "externalDeployment"
            ]["url"]
            == "mock://mock.example.org"
        )

    # Exiting the context manager should delete the last revision, but there's no revision to restore.
    assert (
        mock_graphql["DeleteRevision"].call_args[1]["variable_values"]["revisionId"]
        == "1"
    )
    assert mock_console["SetCurrentRevision"].call_count == 0


@pytest.mark.parametrize(
    "command",
    [
        ["agent", "serve"],
        ["agent", "serve", "foo/bar/baz"],
        ["agent", "deploy", "--no-validate"],
        ["agent", "deploy", "foo/bar/baz", "--no-validate"],
        # venv provisioning takes a long time, so only test it once per command.
        ["agent", "serve", "foo/bar/baz", "--venv"],
        ["agent", "deploy", "foo/bar/baz", "--validate"],
    ],
)
def test_implicit_init(mocker, command, mock_server_process_ctx):
    runner = click.testing.CliRunner()
    with runner.isolated_filesystem():
        mock_server_process_ctx.stop_immediately = True

        result = runner.invoke(
            cli.fixie,
            command,
        )
        assert result.exit_code == 0, result.stdout


def test_fixie_serve_reload(mocker, mock_server_process_ctx):
    runner = click.testing.CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(cli.fixie, ["agent", "init"])
        assert result.exit_code == 0, result.stdout

        # Wait a bit so that filesystem notifications from agent init don't appear as modifications.
        time.sleep(1)

        # Start the server.
        def _run_in_background():
            nonlocal result
            result = runner.invoke(cli.fixie, ["agent", "serve", "--reload"])

        background_thread = threading.Thread(target=_run_in_background, daemon=True)
        background_thread.start()

        # Wait for the server process to be "started".
        stop_server = mock_server_process_ctx.stop_queue.get(timeout=5.0)

        # It should not restart until we modify main.py.
        with pytest.raises(queue.Empty):
            _ = mock_server_process_ctx.stop_queue.get(timeout=0.5)

        # Modify `main.py` and we should see the process get restarted.
        pathlib.Path("main.py").touch()
        stop_server = mock_server_process_ctx.stop_queue.get(timeout=5.0)
        mock_server_process_ctx.__exit__.assert_called_once()

        # Simulate the server process ending.
        stop_server(0)
        background_thread.join()
        assert result.exit_code == 0, result.stdout


def test_fixie_serve_tunnel_rotate(mock_graphql, mock_server_process_ctx, mock_tunnel):
    # Mock the tunnel iterator with a blocking queue.
    tunnel_queue: Optional[queue.Queue] = queue.Queue()

    def _mock_tunnel_iter(*_, **__):
        while tunnel_queue:
            yield tunnel_queue.get()

    mock_tunnel.return_value.__enter__.side_effect = _mock_tunnel_iter

    # Mock revision creation
    mock_revisions_queue: queue.Queue = queue.Queue()

    def _mock_revision_creation(*args, **kwargs):
        mock_revisions_queue.put((args, kwargs))
        return {"createAgentRevision": {"revision": {"id": "FAKEID"}}}

    mock_graphql["CreateAgentRevision"] = _mock_revision_creation

    runner = click.testing.CliRunner()
    with runner.isolated_filesystem():
        result: Optional[click.testing.Result] = None

        # Start the server.
        def _run_in_background():
            nonlocal result
            result = runner.invoke(cli.fixie, ["agent", "serve"])
            assert result.exit_code == 0, result.stdout

        background_thread = threading.Thread(target=_run_in_background, daemon=True)
        background_thread.start()

        # Simulate the initial tunnel URL.
        assert tunnel_queue
        tunnel_queue.put("mock://tunnel1")

        # Wait for the initial revision to be created.
        (_, kwargs) = mock_revisions_queue.get(timeout=5.0)
        assert (
            kwargs["variable_values"]["externalDeployment"]["url"] == "mock://tunnel1"
        )

        # Rotate the tunnel and we should see a second revision created.
        assert tunnel_queue
        tunnel_queue.put("mock://tunnel2")
        (_, kwargs) = mock_revisions_queue.get(timeout=5.0)
        assert (
            kwargs["variable_values"]["externalDeployment"]["url"] == "mock://tunnel2"
        )

        # Simulate the server process ending.
        stop_server = mock_server_process_ctx.stop_queue.get(5.0)
        stop_server(0)

        # Close the mock tunnel.
        assert tunnel_queue
        tunnel_queue.put(None)
        tunnel_queue = None

        background_thread.join()
        assert result
        assert result.exit_code == 0, result.stdout



import dataclasses
from typing import Optional

from . import utils


def test_yaml_formatting():
    @dataclasses.dataclass
    class C(utils.DataClassYamlMixin):
        string: str = "default value"
        optional: Optional[str] = None

    defaults = C()
    assert defaults.to_yaml() == "string: default value\n"

    with_optional = C(optional="optional value")
    assert (
        with_optional.to_yaml() == "string: default value\noptional: optional value\n"
    )

    multiline = C(string="line1\nline2\n", optional="optional value")
    assert (
        multiline.to_yaml() == "string: |\n  line1\n  line2\noptional: optional value\n"
    )



from __future__ import annotations

import logging
from typing import Optional

import jwt
from pydantic import dataclasses as pydantic_dataclasses

from fixieai import constants

# The JWT claim containing the agent ID
_AGENT_ID_JWT_CLAIM = "aid"
_USER_ID_JWT_CLAIM = "sub"

logger = logging.getLogger(__file__)


@pydantic_dataclasses.dataclass
class VerifiedTokenClaims:
    """Verified claims from an agent token."""

    # The agent_id for which the request is intended.
    agent_id: str

    # Indicates whether the request is on behalf of a user or anonymous.
    is_anonymous: bool

    # The verified token.
    token: str

    @staticmethod
    def from_token(
        token: str,
        jwks_client: jwt.PyJWKClient,
        expected_agent_id: Optional[str],
    ) -> Optional[VerifiedTokenClaims]:
        try:
            public_key = jwks_client.get_signing_key_from_jwt(token)
            claims = jwt.decode(
                token,
                public_key.key,
                algorithms=["EdDSA"],
                audience=constants.FIXIE_AGENT_API_AUDIENCES,
            )
        except jwt.DecodeError:
            return None

        token_agent_id = claims.get(_AGENT_ID_JWT_CLAIM)
        if not isinstance(token_agent_id, str):
            # Agent id claim is required
            logger.warning("Rejecting valid JWT without any agent ID claim")
            return None

        if expected_agent_id is not None and token_agent_id != expected_agent_id:
            # The agent ID in the token did not match the allowed value.
            logger.warning(
                f"Rejecting valid JWT because agent ID in token ({token_agent_id!r}) did not match {expected_agent_id!r}"
            )
            return None

        return VerifiedTokenClaims(
            agent_id=token_agent_id,
            is_anonymous=claims.get(_USER_ID_JWT_CLAIM) is None,
            token=token,
        )
